#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <png.h>

/* mkdir */
#ifdef WIN32
#  include <io.h>
#else
#  include <sys/stat.h>
#endif

#include "thanm.h"

/* 0 no output
 * 1 filenames
 * 2 debugging
 * 3 more of it */
static unsigned int verbose = 0;

static int
thtx_to_png(thanm_image_part_t* part, uint16_t format, const char* outname)
{
	FILE *fp;
	png_structp	png_ptr;
	png_infop info_ptr;
	png_byte **image;
	png_byte *row;
	int i, j;

	fp = fopen(outname, "wb");
	if(fp == NULL) {
		perror("fopen");
		return 1;
	}
	png_ptr  = png_create_write_struct(
		PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	info_ptr = png_create_info_struct(png_ptr);
	png_init_io(png_ptr, fp);

	png_color_8 sig_bit = {0};
	switch(format) {
	case THANM_BGRA8888:
		png_set_IHDR(png_ptr, info_ptr, part->w, part->h,
			8, PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE,
			PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
		png_set_bgr(png_ptr);
		png_write_info(png_ptr, info_ptr);

		image = malloc(sizeof(png_bytep) * part->h);
		for(i = 0; i < part->h; ++i) 
			image[i] = part->data + i * part->w * 4;
		png_write_image(png_ptr, image);
		free(image);
		break;
	case THANM_BGR565:
		sig_bit.blue = 5;
		sig_bit.green = 6;
		sig_bit.red = 5;
		png_set_IHDR(png_ptr, info_ptr, part->w, part->h,
			8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
			PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
		png_set_sBIT(png_ptr, info_ptr, &sig_bit);
		png_write_info(png_ptr, info_ptr);
		row = malloc(sizeof(png_bytep) * part->w * 3);
		for(i = 0; i < part->h; ++i) {
			for(j = 0; j < part->w; ++j) {
				unsigned char *p = part->data + (i * part->w + j) * 2;
				row[j * 3 + 0] = p[1] & 0xF8;
				row[j * 3 + 1] = (p[1] << 5) | ((p[0] & 0xe0) >> 3);
				row[j * 3 + 2] = p[0] << 3;
			}
			png_write_row(png_ptr, row);
		}
		free(row);
		break;
	case THANM_BGRA4444: 
		sig_bit.red = sig_bit.green = sig_bit.blue = sig_bit.alpha = 4;
		png_set_IHDR(png_ptr, info_ptr, part->w, part->h,
			8, PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE,
			PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
		png_set_sBIT(png_ptr, info_ptr, &sig_bit);
		png_write_info(png_ptr, info_ptr);
		row = malloc(sizeof(png_bytep) * part->w * 4);
		for(i = 0; i < part->h; ++i) {
			for(j = 0; j < part->w; ++j) {
				unsigned char *p = part->data + (i * part->w + j) * 2;
				row[j * 4 + 0] = p[1] << 4;
				row[j * 4 + 1] = p[0] & 0xF0;
				row[j * 4 + 2] = p[0] << 4;
				row[j * 4 + 3] = p[1] & 0xF0;
			}
			png_write_row(png_ptr, row);
		}
		free(row);
		break;
	default:
		fprintf(stderr, "anm2png: unknown pixelformat ``%u''\n", format);
		return 1;
	}

	png_write_end(png_ptr, info_ptr);
	png_destroy_write_struct(&png_ptr, &info_ptr);
	fclose(fp);
	return 0;
}

static void
join_anm_parts(thanm_image_part_t* parts, unsigned int count, thanm_image_part_t* out, unsigned int Bpp)
{
	unsigned int i, j;

	out->x = out->y = out->w = out->h = 0;
	out->data = NULL;

	for (i = 0; i < count; ++i) {
		if (parts[i].x + parts[i].w > out->w)
			out->w = parts[i].x + parts[i].w;
		if (parts[i].y + parts[i].h > out->h)
			out->h = parts[i].y + parts[i].h;
	}

	out->data = malloc(out->w * Bpp * out->h);

	for (i = 0; i < count; ++i) {
		for (j = 0; j < parts[i].h; ++j) {
			memcpy(out->data + ((parts[i].y + j) * out->w + parts[i].x) * Bpp, parts[i].data + j * parts[i].w * Bpp, parts[i].w * Bpp);
		}
	}
}

static int
thanm_extract_file(thanm_t* anm, const char* path)
{
	unsigned int i = 0, j = 0;
	uint16_t format = 0;

	if (verbose >= 1)
		printf("%s\n", path);

	do {
		if (strcmp(anm->entries[i].name, path) == 0) {
			format = anm->entries[i].header.format;
			j++;
		}
	} while(anm->entries[i++].header.nextoffset);

	if (j == 0) {
		fprintf(stderr, "anm2png: path not found\n");
		return -1;
	}

	thanm_image_part_t* parts = malloc(j * sizeof(thanm_image_part_t));

	i = j = 0;
	do {
		if (strcmp(anm->entries[i].name, path) == 0) {
			if (verbose >= 1 && format != anm->entries[i].header.format) {
				fprintf(stderr, "anm2png: format mismatch (%u, %u)\n", format, anm->entries[i].header.format);
			}

			if (fseek(anm->stream, anm->entries[i].offset + anm->entries[i].header.entryoffset, SEEK_SET) == -1) {
				fprintf(stderr, "anm2png: couldn't seek: %s\n", strerror(errno));
				return -1;
			}

			thtx_header_t txheader;
			/* sizeof(thtx_header_t) is 0x10 */
			if (fread(&txheader, 0x10, 1, anm->stream) != 1) {
				fprintf(stderr, "anm2png: couldn't read %u bytes: %s\n", 0x10, strerror(errno));
				return -1;
			}

			parts[j].x = anm->entries[i].header.offsetx;
			parts[j].y = anm->entries[i].header.offsety;
			parts[j].w = txheader.width;
			parts[j].h = txheader.height;
			parts[j].data = malloc(txheader.size);
			if (!parts[j].data) {
				fprintf(stderr, "anm2png: malloc failed: %s\n", strerror(errno));
				return -1;
			}
			if (fread(parts[j].data, txheader.size, 1, anm->stream) != 1) {
				fprintf(stderr, "anm2png: couldn't read %u bytes: %s\n", txheader.size, strerror(errno));
				return -1;
			}
			j++;
		}
	} while(anm->entries[i++].header.nextoffset);

	thanm_image_part_t out;
	if (j == 1)
		out = parts[0];
	else
		join_anm_parts(parts, j, &out, format == THANM_BGRA8888 ? 4 : 2);

	char* name = strdup(path);
	char* filename = name;
	for (;;) {
		/* The +1 is not correct for the first pass, but it prevents
		 * creating files anywhere on the filesystem while simplifying
		 * the code. */
		filename = strchr(filename + 1, '/');
		if (!filename)
			break;
		*filename = '\0';
		/* Ignore any errors... */
#ifdef WIN32
		mkdir(name);
#else
		mkdir(name, 0755);
#endif
		*filename = '/';
	}
	free(name);

	thtx_to_png(&out, format, path);

	free(out.data);

	if (j != 1)
		for (i = 0; i < j; ++i)
			free(parts[i].data);
	free(parts);

	return 0;
}

static int
thanm_extract(thanm_t* anm)
{
	unsigned int i = 0;
	do {
		unsigned int j;
		/* This one appears to be zero for invalid entries. */
		if (anm->entries[i].header.unknown2 == 0)
			continue;
		for (j = 0; j < i; ++j) {
			if (strcmp(anm->entries[i].name, anm->entries[j].name) == 0)
				break;
		}
		if (i == j) {
			if (thanm_extract_file(anm, anm->entries[i].name) == -1)
				return -1;
		}
	} while(anm->entries[i++].header.nextoffset);

	return 0;
}

int
main(int argc, char* argv[])
{
	unsigned int i;
	unsigned int ignoreopts = 0;
	unsigned int anmid = 0; /* argument index of anm archive */
	unsigned int pathid = 0; /* argument index of list of paths */
	thanm_t* anm;
	for (i = 1; i < argc; ++i) {
		if (anmid == 0) {
			if (argv[i][0] == '-' && !ignoreopts) {
				switch (argv[i][1]) {
				case '-':
					ignoreopts = 1;
					break;
				case 'v':
					verbose += strlen(argv[i]) - 1;
					break;
				default:
					fprintf(stderr, "anm2png: unrecognized option: %s\n", argv[i]);
					break;
				}
			} else {
				anmid = i;
			}
		} else if (pathid == 0) {
			pathid = i;
			break;
		}
	}

	if (!anmid) {
		fprintf(stderr,
			"Usage: anm2png [OPTIONS] FILE.anm [PATH...]\n"
			"Where FILE.anm is an image archive and PATH is a list of images to extract (default to all).\n"
			"OPTIONS can be -v for verbose operation. Use several times for increasingly operation.\n");
		return EXIT_FAILURE;
	}

	anm = thanm_open(argv[anmid]);
	if (!anm) {
		fprintf(stderr, "anm2png: couldn't open ``%s''\n", argv[anmid]);
		return EXIT_FAILURE;
	}

	if (pathid) {
		for (i = pathid; i < argc; ++i) {
			if (thanm_extract_file(anm, argv[i]) == -1) {
				fprintf(stderr, "anm2png: couldn't extract ``%s'' from ``%s''\n", argv[i], argv[anmid]);
				thanm_close(anm);
				return EXIT_FAILURE;
			}
		}
	} else {
		if (thanm_extract(anm) == -1) {
			fprintf(stderr, "anm2png: couldn't extract from ``%s''\n", argv[anmid]);
			thanm_close(anm);
			return EXIT_FAILURE;
		}
	}
	thanm_close(anm);

	return EXIT_SUCCESS;
}
