#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include "thanm.h"

typedef struct
{
	uint32_t regnum;
	uint32_t attrnum;
	uint32_t zero1;
	uint32_t width;
	uint32_t height;
	uint32_t format;
	uint32_t zero2;
	uint32_t nameoffset;
	uint32_t zero3;
	uint32_t zero4;
	uint32_t version;
	uint32_t unknown1; /* 10 (0, 11?) */
	uint32_t entryoffset;
	uint32_t unknown2; /* 0, 1 */
	uint32_t nextoffset;
	uint32_t zero5;
} thanm_header_th10_t;

/* Convert a TH10 header into a TH11 header by reordering some fields. */
static void
convert_header(thanm_header_t* header)
{
	thanm_header_th10_t th10 = *(thanm_header_th10_t*)header;
	header->regnum = th10.regnum;
	header->attrnum = th10.attrnum;
	header->width = th10.width;
	header->height = th10.height;
	header->format = th10.format;
	header->nameoffset = th10.nameoffset;
	header->version = th10.version;
	header->entryoffset = th10.entryoffset;
	header->nextoffset = th10.nextoffset;
	header->offsetx = 0;
	header->offsety = 0;

	header->unknown1 = th10.unknown1;
	header->unknown2 = th10.unknown2;
}

/* Read a stream until '\0'. */
static ssize_t
read_asciiz(FILE* stream, unsigned int buffersize, char* buffer)
{
	unsigned int i;
	int c;
	for (i = 0; i < buffersize; ++i) {
		if (i == buffersize - 1) {
			return -1;
		}
		c = fgetc(stream);
		if (c == EOF) {
			return -1;
		}
		buffer[i] = c;
		if (!c)
			break;
	}
	return i + 1;
}

thanm_t*
thanm_open(const char* filename)
{
	thanm_t* anm = malloc(sizeof(thanm_t));
	anm->stream = fopen(filename, "rb");
	if (!anm->stream) {
		free(anm);
		return NULL;
	}

	unsigned int offset = 0;
	unsigned int i = 0;

	anm->entries = NULL;

	do {
		anm->entries = realloc(anm->entries, (i + 1) * sizeof(thanm_entry_t));
		if (!anm->entries) {
			/* XXX: Leaks anm->entries if NULL is returned. */
			free(anm);
			return NULL;
		}
		if (fseek(anm->stream, offset, SEEK_SET) == -1) {
			free(anm->entries);
			free(anm);
			return NULL;
		}
		if (fread(&anm->entries[i].header, sizeof(thanm_header_t), 1, anm->stream) != 1) {
			free(anm->entries);
			free(anm);
			return NULL;
		}
		if (anm->entries[i].header.version != 7) {
			convert_header(&anm->entries[i].header);
			if (anm->entries[i].header.version != 4) {
				free(anm->entries);
				free(anm);
				return NULL;
			}
		}

		if (fseek(anm->stream, offset + anm->entries[i].header.nameoffset, SEEK_SET) == -1) {
			free(anm->entries);
			free(anm);
			return NULL;
		}

		char buffer[PATH_MAX];
		ssize_t length = read_asciiz(anm->stream, PATH_MAX, buffer);
		if (!length) {
			free(anm->entries);
			free(anm);
			return NULL;
		}

		anm->entries[i].name = strdup(buffer);
		anm->entries[i].offset = offset;
		
		offset += anm->entries[i].header.nextoffset;
	} while(anm->entries[i++].header.nextoffset);

	return anm;
}

void
thanm_close(thanm_t* anm)
{
	unsigned int i = 0;
	fclose(anm->stream);
	do {
		free(anm->entries[i].name);
	} while(anm->entries[i++].header.nextoffset);
	free(anm->entries);
	free(anm);
}
