/* TODO: Make the program work on big-endian architectures. */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <png.h>

#include "thanm.h"

static int
thinspng_loadpng(const char* path, unsigned int format, thanm_image_part_t* out)
{
	unsigned char header[8];
	FILE* pngfd;
    png_structp png_ptr;
    png_infop info_ptr;
	png_bytepp row_pointers;
	unsigned int i, j;
	uint16_t* data16;

	pngfd = fopen(path, "rb");
	if (!pngfd) {
		fprintf(stderr, "thinspng: couldn't open ``%s'' for reading: %s\n",
			path, strerror(errno));
		return -1;
	}

	if (fread(header, 8, 1, pngfd) != 1) {
		fprintf(stderr, "thinspng: couldn't read header from ``%s'': %s\n",
			path, strerror(errno));
		fclose(pngfd);
		return -1;
	}

	if (png_sig_cmp(header, 0, 8) != 0) {
		fprintf(stderr, "thinspng: ``%s'' does not appear to be a PNG file\n",
			path);
		fclose(pngfd);
		return -1;
	}

    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr) {
		fprintf(stderr, "thinspng: couldn't create PNG read struct\n");
		fclose(pngfd);
        return -1;
	}

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr) {
		fprintf(stderr, "thinspng: couldn't create PNG info struct\n");
		fclose(pngfd);
        png_destroy_read_struct(&png_ptr, NULL, NULL);
        return -1;
    }

	png_init_io(png_ptr, pngfd);
	png_set_sig_bytes(png_ptr, 8);

	/* PNG_TRANSFORM_SHIFT */
	png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_BGR, NULL);

	if (out->w != info_ptr->width || out->h != info_ptr->height) {
		fprintf(stderr, "thinspng: PNG image dimensions (%u, %u) different from ANM (%u, %u)\n",
			(unsigned int)info_ptr->width, (unsigned int)info_ptr->height, out->w, out->h);
		fclose(pngfd);
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		return -1;
	}

	if ((format == THANM_BGRA8888 || format == THANM_BGRA4444) && info_ptr->color_type != PNG_COLOR_TYPE_RGB_ALPHA) {
		fprintf(stderr, "thinspng: ``%s'' must be RGBA\n", path);
		fclose(pngfd);
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		return -1;
	} else if (format == THANM_BGR565 && info_ptr->color_type != PNG_COLOR_TYPE_RGB) {
		fprintf(stderr, "thinspng: ``%s'' must be RGB\n", path);
		fclose(pngfd);
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		return -1;
	}

	row_pointers = png_get_rows(png_ptr, info_ptr);
	fclose(pngfd);

	for (i = 0; i < out->h; ++i) {
		if (format == THANM_BGRA8888) {
			memcpy(out->data + (i * out->w * 4), row_pointers[i], out->w * 4);
		} else if (format == THANM_BGR565) {
			for (j = 0; j < out->w; j++) {
				data16 = (uint16_t*)(out->data + 2 * (i * out->w + j));
				*data16 = ((row_pointers[i][j * 3] >> 3))
					    | ((row_pointers[i][j * 3 + 1] >> 2) << 5)
						| ((row_pointers[i][j * 3 + 2] >> 3) << 11);
			}
		} else if (format == THANM_BGRA4444) {
			for (j = 0; j < out->w; j++) {
				out->data[i * 2 * out->w + (j * 2)]     = (row_pointers[i][j * 4 + 0] >> 4) | ((row_pointers[i][j * 4 + 1] >> 4) << 4);
				out->data[i * 2 * out->w + (j * 2) + 1] = (row_pointers[i][j * 4 + 2] >> 4) | ((row_pointers[i][j * 4 + 3] >> 4) << 4);
			}
		} else {
			fprintf(stderr, "thinspng: unknown ANM format: %u\n", format);
			return -1;
		}
	}

	png_destroy_read_struct(&png_ptr, &info_ptr, NULL);

	return 0;
}
	
int
main(int argc, char* argv[])
{
	FILE* anmfd;
	unsigned int i, j;
	unsigned int format = 0;
	int parts_count = 0;
	unsigned int Bpp;

	if (argc != 4) {
		fprintf(stderr, "Usage: thinspng FILE.anm PATH FILE.png\nReplaces the PATH entry in FILE.anm by FILE.png.\n");
		return EXIT_FAILURE;
	}

	thanm_t* anm;
	anm = thanm_open(argv[1]);
	if (!anm) {
		fprintf(stderr, "thinspng: couldn't open ``%s''\n", argv[1]);
		return EXIT_FAILURE;
	}

	/* find all the parts */
	i = 0;
	j = 0;
	do {
		if (strcmp(anm->entries[i].name, argv[2]) == 0) {
			format = anm->entries[i].header.format;
			j++;
		}
	} while(anm->entries[i++].header.nextoffset);

	if (j == 0) {
		fprintf(stderr, "thinspng: couldn't find ``%s'' in ``%s''\n",
			argv[2], argv[1]);
		thanm_close(anm);
		return EXIT_FAILURE;
	}

	thanm_image_part_t* parts = malloc(j * sizeof(thanm_image_part_t));
	parts_count = j;

	i = j = 0;
	do {
		if (strcmp(anm->entries[i].name, argv[2]) == 0) {
			if (format != anm->entries[i].header.format) {
				fprintf(stderr, "thinspng: format mismatch\n");
				free(parts);
				thanm_close(anm);
				return EXIT_FAILURE;
			}

			if (fseek(anm->stream, anm->entries[i].offset + anm->entries[i].header.entryoffset, SEEK_SET) == -1) {
				fprintf(stderr, "thinspng: couldn't seek: %s", strerror(errno));
				free(parts);
				thanm_close(anm);
				return EXIT_FAILURE;
			}

			thtx_header_t txheader;
			if (fread(&txheader, 0x10, 1, anm->stream) != 1) {
				fprintf(stderr, "thinspng: couldn't read %u bytes: %s", 0x10, strerror(errno));
				free(parts);
				thanm_close(anm);
				return EXIT_FAILURE;
			}

			parts[j].x = anm->entries[i].header.offsetx;
			parts[j].y = anm->entries[i].header.offsety;
			parts[j].w = txheader.width;
			parts[j].h = txheader.height;
			parts[j].data = malloc(txheader.size);
			parts[j].offset = anm->entries[i].offset + anm->entries[i].header.entryoffset + sizeof(thtx_header_t);
			j++;
		}
	} while(anm->entries[i++].header.nextoffset);

	Bpp = format == THANM_BGRA8888 ? 4 : 2;

	thanm_image_part_t out;
	out.x = out.y = out.w = out.h = 0;
	out.data = NULL;

	for (i = 0; i < parts_count; ++i) {
		if (parts[i].x + parts[i].w > out.w)
			out.w = parts[i].x + parts[i].w;
		if (parts[i].y + parts[i].h > out.h)
			out.h = parts[i].y + parts[i].h;
	}

	out.data = malloc(out.w * Bpp * out.h);
	if (!out.data) {
		fprintf(stderr, "thinspng: malloc failed: %s\n", strerror(errno));
		free(parts);
		thanm_close(anm);
		return EXIT_FAILURE;
	}

	if (thinspng_loadpng(argv[3], format, &out) != 0) {
		fprintf(stderr, "thinspng: couldn't load image from ``%s''\n",
			argv[3]);
		thanm_close(anm);
		return EXIT_FAILURE;
	}

	for (i = 0; i < parts_count; ++i) {
		for (j = 0; j < parts[i].h; ++j) {
			memcpy(parts[i].data + parts[i].w * j * Bpp,
			       out.data + (out.w * (parts[i].y + j) * Bpp) + (parts[i].x * Bpp),
			       parts[i].w * Bpp);
		}
	}

	free(out.data);

	thanm_close(anm);

	/* Write new data to the file. */
	anmfd = fopen(argv[1], "rb+");
	if (!anmfd) {
		fprintf(stderr, "thinspng: couldn't open ``%s'' for writing: %s\n",
			argv[1], strerror(errno));
		return EXIT_FAILURE;
	}

	for (i = 0; i < parts_count; ++i) {
		if (fseek(anmfd, parts[i].offset, SEEK_SET) == -1) {
			fprintf(stderr, "thinspng: couldn't seek to %u in ``%s'': %s\n",
				parts[i].offset, argv[1], strerror(errno));
			fclose(anmfd);
			return EXIT_FAILURE;
		}
		if (fwrite(parts[i].data, parts[i].w * parts[i].h * Bpp, 1, anmfd) != 1) {
			fprintf(stderr, "thinspng: couldn't write to ``%s'': %s\n",
				argv[1], strerror(errno));
			fclose(anmfd);
			return EXIT_FAILURE;
		}
		free(parts[i].data);
	}

	free(parts);
	fclose(anmfd);

	return EXIT_SUCCESS;
}
